package edu.illinois.cs.cs124.ay2024.mp.test

import android.content.Intent
import androidx.test.core.app.ApplicationProvider
import androidx.test.espresso.Espresso.onView
import androidx.test.espresso.action.ViewActions.click
import androidx.test.espresso.assertion.ViewAssertions.matches
import androidx.test.espresso.matcher.ViewMatchers.hasDescendant
import androidx.test.espresso.matcher.ViewMatchers.isDisplayed
import androidx.test.espresso.matcher.ViewMatchers.withId
import androidx.test.espresso.matcher.ViewMatchers.withText
import androidx.test.ext.junit.runners.AndroidJUnit4
import com.google.common.truth.Truth.assertWithMessage
import edu.illinois.cs.cs124.ay2024.mp.R
import edu.illinois.cs.cs124.ay2024.mp.activities.RSOActivity
import edu.illinois.cs.cs124.ay2024.mp.models.RSO
import edu.illinois.cs.cs124.ay2024.mp.models.Summary
import edu.illinois.cs.cs124.ay2024.mp.network.Client
import edu.illinois.cs.cs124.ay2024.mp.network.startServer
import edu.illinois.cs.cs124.ay2024.mp.test.helpers.GET_METHOD_EXTRA_TIME
import edu.illinois.cs.cs124.ay2024.mp.test.helpers.JSONReadCountSecurityManager
import edu.illinois.cs.cs124.ay2024.mp.test.helpers.RSODATA
import edu.illinois.cs.cs124.ay2024.mp.test.helpers.SUMMARIES
import edu.illinois.cs.cs124.ay2024.mp.test.helpers.SUMMARY_COUNT
import edu.illinois.cs.cs124.ay2024.mp.test.helpers.configureLogging
import edu.illinois.cs.cs124.ay2024.mp.test.helpers.countRecyclerView
import edu.illinois.cs.cs124.ay2024.mp.test.helpers.doCheckServerDesign
import edu.illinois.cs.cs124.ay2024.mp.test.helpers.getShuffledSummaries
import edu.illinois.cs.cs124.ay2024.mp.test.helpers.pause
import edu.illinois.cs.cs124.ay2024.mp.test.helpers.startActivity
import edu.illinois.cs.cs124.ay2024.mp.test.helpers.startMainActivity
import edu.illinois.cs.cs124.ay2024.mp.test.helpers.testClient
import edu.illinois.cs.cs124.ay2024.mp.test.helpers.testServerGet
import edu.illinois.cs.cs124.ay2024.mp.test.helpers.testServerGetTimed
import edu.illinois.cs.cs124.ay2024.mp.test.helpers.testSummaryRoute
import edu.illinois.cs.cs124.ay2024.mp.test.helpers.trimmedMean
import edu.illinois.cs.cs124.ay2024.mp.test.helpers.withRecyclerView
import edu.illinois.cs.cs125.gradlegrader.annotations.Graded
import org.hamcrest.BaseMatcher
import org.hamcrest.Description
import org.hamcrest.Matcher
import org.hamcrest.Matchers.containsString
import org.junit.After
import org.junit.AfterClass
import org.junit.Before
import org.junit.BeforeClass
import org.junit.FixMethodOrder
import org.junit.Test
import org.junit.runner.RunWith
import org.junit.runners.MethodSorters
import org.robolectric.Shadows.shadowOf
import org.robolectric.annotation.LooperMode
import org.robolectric.annotation.experimental.LazyApplication
import java.net.HttpURLConnection

/*
 * This is the MP2 test suite.
 * The code below is used to evaluate your app during testing, local grading, and official grading.
 * You may not understand all of the code below, but you'll need to have some understanding of how
 * it works so that you can determine what is wrong with your app and what you need to fix.
 *
 * ALL CHANGES TO THIS FILE WILL BE OVERWRITTEN DURING OFFICIAL GRADING.
 * You can and should modify the code below if it is useful during your own local testing,
 * but any changes you make will be discarded during official grading.
 * The local grader will not run if the test suites have been modified, so you'll need to undo any
 * local changes before you run the grader.
 *
 * Note that this means that you should not fix problems with the app by modifying the test suites.
 * The test suites are always considered to be correct.
 *
 * Our test suites are broken into two parts.
 * The unit tests are tests that we can perform without running your app.
 * They test things like whether a method works properly or the behavior of your API server.
 * Unit tests are usually fairly fast.
 *
 * The integration tests are tests that require simulating your app.
 * This allows us to test things like your API client, and higher-level aspects of your app's
 * behavior, such as whether it displays the right thing on the display.
 * Because integration tests require simulating your app, they run more slowly.
 *
 * The MP2 test suite includes no ungraded tests.
 * These tests are fairly idiomatic, in that they resemble tests you might write for an actual
 * Android programming project.
 */

@RunWith(AndroidJUnit4::class)
@LooperMode(LooperMode.Mode.PAUSED)
@FixMethodOrder(MethodSorters.NAME_ASCENDING)
class MP2Test {
    // Test the RSO class design
    @Test(timeout = 4000L)
    @Graded(points = 10, friendlyName = "RSO Class Design (Unit)")
    @LazyApplication(LazyApplication.LazyLoad.ON)
    fun test0_RSOClassDesign() {
        // Allows us to count RSO categories, non-empty mission statements, and external websites,
        // ensuring that the RSO class is designed properly and parsing the input data correctly
        val counts = RSOCounts()

        // Initialize all RSO classes using RSOData objects
        for (rsoData in RSODATA) {
            val rso = RSO(rsoData)
            // Add the new RSO to our counts
            counts.countRSO(rso)
        }

        // Check that the counts match what we expect
        counts.checkCounts(2059, 1152, 53)
    }

    // Test GET /rso server route
    @Test(timeout = 60000L)
    @Graded(points = 20, friendlyName = "Server GET /rso (Unit)")
    @LazyApplication(LazyApplication.LazyLoad.ON)
    fun test1_ServerRSORoute() {
        // Test the /summary route first as a sanity check
        testSummaryRoute()

        // Save response times for comparison
        val baseResponseTimes = mutableListOf<Long>()
        val getResponseTimes = mutableListOf<Long>()

        // Get a random subset of all RSO summaries
        val trimmedSummaries: List<Summary> = getShuffledSummaries(12421, 128)

        // Allows us to count RSO categories, non-empty mission statements, and external websites,
        // ensuring that the RSO results returned by the server are correct
        val counts = RSOCounts()

        // Round 0: Test good GET /rso requests for all RSOs
        // Untimed to warm up the server and not fail after MP3
        for (rsoSummary in trimmedSummaries) {
            // Perform the main RSO GET
            val rso = "/rso/${rsoSummary.id}".testServerGet<RSO>(RSO::class.java)

            // RSO returned should not be null
            assertWithMessage("RSO returned from server GET /rso should not be null")
                .that(rso)
                .isNotNull()

            // Count the returned RSO object
            counts.countRSO(rso!!)
        }

        // Check that the counts match what we expect
        counts.checkCounts(222, 120, 3)

        // Clear the counts for the timed round
        counts.clear()

        // Round 1: Test good GET /rso requests for all RSOs
        // Timed to detect slow routes
        for (rsoSummary in trimmedSummaries) {
            // Time the index route for comparison
            val baseResult = "/".testServerGetTimed<String>(String::class.java)
            baseResponseTimes.add(baseResult.responseTime.toNanos())

            // Perform the main RSO GET
            val rsoResult = "/rso/${rsoSummary.id}".testServerGetTimed<RSO>(RSO::class.java)
            getResponseTimes.add(rsoResult.responseTime.toNanos())

            // RSO returned should not be null
            assertWithMessage("RSO returned from server GET /rso should not be null")
                .that(rsoResult.response)
                .isNotNull()

            // Count the returned RSO object
            counts.countRSO(rsoResult.response!!)
        }

        // Check that the counts match what we expect
        counts.checkCounts(222, 120, 3)

        // Check for slow server GETs potentially caused by unnecessary parsing or looping
        val averageBase = trimmedMean(baseResponseTimes, 0.1)
        val averageResponse = trimmedMean(getResponseTimes, 0.1)
        assertWithMessage("Server GET /rso is too slow")
            .that(averageResponse)
            .isLessThan(averageBase * GET_METHOD_EXTRA_TIME)

        // Test bad requests
        "/rsos/KUv_oPtSyVms6QUnd2VzZu_3CVI/".testServerGet(HttpURLConnection.HTTP_NOT_FOUND)
        // Non-existent RSO
        "/rso/KUv_oPtSyVms6QUnd2VZZu_3CVI".testServerGet(HttpURLConnection.HTTP_NOT_FOUND)
        // Non-existent URL
        "/rsos/KUv_oPtSyVms6QUnd2VzZu_3CVI/".testServerGet(HttpURLConnection.HTTP_NOT_FOUND)
    }

    // Test the Client getRSO method
    @Test(timeout = 40000L)
    @Graded(points = 20, friendlyName = "Client getRSO (Integration)")
    @LazyApplication(LazyApplication.LazyLoad.ON)
    fun test2_ClientGetRSO() {
        // Get a random subset of all RSO summaries
        val trimmedSummaries = getShuffledSummaries(12422, 128)

        // Allows us to count RSO categories, non-empty mission statements, and external websites,
        // ensuring that the RSO results returned by the client are correct
        val counts = RSOCounts()

        // Use the Client to make RSO requests for all RSO summaries in our subset
        for (summary in trimmedSummaries) {
            val rso = testClient { callback -> Client.getRSO(summary.id, callback) }
            // Check that the returned RSO id is correct
            assertWithMessage("Incorrect id on returned RSO")
                .that(rso.id)
                .isEqualTo(summary.id)
            // Count the returned RSO object
            counts.countRSO(rso)
        }

        // Check that the counts match what we expect
        counts.checkCounts(217, 124, 3)

        // Test bad getRSO request
        try {
            testClient { callback -> Client.getRSO("_6oRAM9nit1fntTdDkIOk_Q2nVS", callback) }
            assertWithMessage("Client GET /rso/ for non-existent RSO should throw").fail()
        } catch (ignored: Exception) {
        }
    }

    // Test onClick intent generation in the main activity
    @Suppress("SpellCheckingInspection")
    @Test(timeout = 20000L)
    @Graded(points = 10, friendlyName = "Summary Click Launch (Integration)")
    @LazyApplication(LazyApplication.LazyLoad.ON)
    fun test3_SummaryClickLaunch() {
        startMainActivity { activity ->
            // Double-check that the number of RSO summaries displayed is correct
            onView(withId(R.id.recycler_view)).check(countRecyclerView(SUMMARY_COUNT))
            // Double-check that the RSO summaries are sorted as we expect
            onView(withRecyclerView(R.id.recycler_view).atPosition(2))
                .check(matches(hasDescendant(withText("8 to CREATE"))))

            // Perform the click
            onView(withRecyclerView(R.id.recycler_view).atPosition(2)).perform(click())

            // Make sure the intent generated is correct
            val id = shadowOf(activity).nextStartedActivity.getStringExtra("id")
            assertWithMessage("Incorrect id in intent")
                .that(id)
                .isEqualTo("eIm2apBazz95gGylU3a8Mbd0vFY")
        }
    }

    // Test RSOActivity UI launched via intent
    @Test(timeout = 20000L)
    @Graded(points = 20, friendlyName = "RSO View (Integration)")
    @LazyApplication(LazyApplication.LazyLoad.ON)
    fun test4_RSOView() {
        // Get a random subset of all RSO summaries
        val trimmedSummaries = getShuffledSummaries(12424, 32)

        // Use the random RSOs we picked above
        // Can't use the for-in loop here since we maintain two counters
        var count = 0
        var i = 0
        while (i < 32 && count < 4) {
            i++
            val summary = trimmedSummaries[i]

            // Create the Intent
            val intent = Intent(ApplicationProvider.getApplicationContext(), RSOActivity::class.java)
            intent.putExtra("id", summary.id)

            // Use the Client to retrieve the RSO details
            val rso = testClient { callback -> Client.getRSO(summary.id, callback) }
            // Skip RSOs without a website, mission statement, or categories
            if (rso.website.isBlank() || rso.mission.isBlank() || rso.categories.isEmpty()) {
                continue
            }
            count++

            // Start the RSO activity
            startActivity<RSOActivity>(intent) {
                // Wait for it to load...
                pause()

                // Check for the title, website, mission statement, and categories
                // Title, website, and mission should appear in their own view
                onView(withText(rso.title)).check(matches(isDisplayed()))
                onView(withText(rso.website)).check(matches(isDisplayed()))
                onView(withText(rso.mission)).check(matches(isDisplayed()))

                // Category strings may eventually appear multiple times, and formatting may vary.
                // So we'll just check that all the categories for this RSO are shown somewhere
                rso.categories.forEach { categoryString ->
                    onView(first(withText(containsString(categoryString))))
                        .check(matches(isDisplayed()))
                }
            }
        }
    }

    companion object {
        // Security manager used to count access to rsos.json
        private val jsonReadCountSecurityManager = JSONReadCountSecurityManager()

        // Run once before any test in this suite is started
        @JvmStatic
        @BeforeClass
        fun beforeClass() {
            // Check Server.kt for publicly visible members
            doCheckServerDesign()
            // Set up logging so that you can see log output during testing
            configureLogging()
            // Force loads to perform initialization before we start counting
            RSODATA.size
            SUMMARIES.size
            // Install our security manager that allows counting access to rsos.json
            System.setSecurityManager(jsonReadCountSecurityManager)
            // Start the API server
            startServer()
        }

        // Run once after all tests in this suite are completed
        @JvmStatic
        @AfterClass
        fun removeSecurityManager() {
            System.setSecurityManager(null)
        }
    }

    // Run before each test
    @Before
    fun beforeEachTest() {
        // Check for extra reads from rsos.json
        jsonReadCountSecurityManager.checkCount()
    }

    // Run after each test
    @After
    fun afterEachTest() {
        // Check for extra reads from rsos.json
        jsonReadCountSecurityManager.checkCount()
    }
}

/**
 * Used to count RSO attributes for testing verification.
 */
data class RSOCounts(
    var totalCategories: Int = 0,
    var totalNonEmptyMissionStatements: Int = 0,
    var totalNonOneIllinoisWebsites: Int = 0,
) {
    fun countRSO(rso: RSO) {
        totalCategories += rso.categories.size
        if (rso.mission.isNotEmpty()) {
            totalNonEmptyMissionStatements += 1
        }
        if (!rso.website.startsWith("https://one.illinois.edu")) {
            totalNonOneIllinoisWebsites += 1
        }
    }

    fun checkCounts(
        expectedTotalCategories: Int,
        expectedTotalNonEmptyMissionStatements: Int,
        expectedTotalNonOneIllinoisWebsites: Int,
    ) {
        assertWithMessage("Incorrect total number of RSO categories")
            .that(totalCategories)
            .isEqualTo(expectedTotalCategories)
        assertWithMessage("Incorrect total number of non-empty mission statements")
            .that(totalNonEmptyMissionStatements)
            .isEqualTo(expectedTotalNonEmptyMissionStatements)
        assertWithMessage("Incorrect total number non-one.illinois.edu websites")
            .that(totalNonOneIllinoisWebsites)
            .isEqualTo(expectedTotalNonOneIllinoisWebsites)
    }

    fun clear() {
        totalCategories = 0
        totalNonEmptyMissionStatements = 0
        totalNonOneIllinoisWebsites = 0
    }
}

// View matcher to select first matching view when multiple match our criteria
private fun <T> first(matcher: Matcher<T>): Matcher<T> {
    return object : BaseMatcher<T>() {
        var isFirst: Boolean = true

        override fun matches(item: Any): Boolean {
            if (isFirst && matcher.matches(item)) {
                isFirst = false
                return true
            }
            return false
        }

        override fun describeTo(description: Description) {
            description.appendText("should return first matching item")
        }
    }
}

// md5: 6a3611cf38a830a033977a5a2a573d2c // DO NOT REMOVE THIS LINE
